--[[
	NOTE: Many parts of this Addon are copied from RCLootCouncil. Credit to the original author(s). 
]]--

SFLoot = LibStub("AceAddon-3.0"):NewAddon("SFLoot", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0", "AceSerializer-3.0", "AceTimer-3.0");
local lwin = LibStub("LibWindow-1.1");
local LibDialog = LibStub("LibDialog-1.0");
local db;

local frames = {};
frames.itemButtons = {};
frames.lootFrames = {};
local menuFrame;
local lootTable = {};
local mlLoot = {};
local currentItem = 1;
local responses = {
	[1] = "Pass",
	[2] = "Transmog",
	[3] = "Off",
	[4] = "Main",
};
local response_colors = {
	[1] = {1, 0, 0, 1}, --red
	[2] = {1, 0.5, 0, 1}, -- orange
	[3] = {0, 0.8, 1, 1}, -- blue
	[4] = {0, 1, 0, 1}, -- green
};

local requested_rolls = {}

local testing = false;

local today = date("%Y/%m/%d", time())

local ROW_HEIGHT = 20;
local NUM_ROWS = 10;

function SFLoot:OnInitialize()

	self.scrollCols = {
		{ name = "",																	width = 20, },	-- 1 Class
		{ name = "Name",																	width = 150,},	-- 2 Candidate Name
		{ name = "Verwendung",						comparesort = ResponseSort, sortnext = 9,		width = 80,},	-- 3 Response
		{ name = "ilvl",															width = 45, },	-- 4 Total ilvl
		{ name = "Diff",																			width = 40, },	-- 5 ilvl difference
		{ name = "g1",			align = "CENTER",								width = 20, },	-- 6 Current gear 1
		{ name = "g2",			align = "CENTER",								width = 20, },	-- 7 Current gear 2
		{ name = "Items",		align = "CENTER",																	width = 40, },	-- 8 num items
		{ name = "Roll",		align = "CENTER", 	comparesort = RollSort,				width = 30, },	-- 9 Roll
	}

	self.db = LibStub("AceDB-3.0"):New("SFLootDB", defaults)
	
	self.defaults = {
		global = {
			localizedSubTypes = {},
		},
		profile = {
			debug =  false,
			ambiguate = false,
			auto_open = true,
			auto_close = false,
			change_loot_method = true,
			secure_roll = false,
			frame_scale = 0.85,
			show_loot_anchor = true,
			loot_frame_direction = false, --true = grow down, false = grow up
			show_on_roll = true,
			lf_always = false,
			lf_main = true,
			lf_off = true,
			lf_transmog = true,
			lf_pass = false,
			UI = { -- stores all ui information
				['**'] = { -- Defaults
					y		= 0,
					x		= 0,
					point	= "CENTER",
					scale	= 0.8,
					bgColor = {0.25, 0.25, 0.25, 1},
					borderColor = {1, 1, 1, 0.2},
					border = "Interface\\Tooltips\\UI-Tooltip-Border",
					background = "Interface\\Tooltips\\UI-Tooltip-Background",
				},
				default = {}, -- base line
			},
			lootDB = {},
			-- List of items to ignore:
			ignore = {
				109693, -- Draenic Dust
				115502, -- Small Luminous Shard
				111245, -- Luminous Shard
				115504, -- Fractured Temporal Crystal
				113588, -- Temporal Crystal
				124442, -- Chaos Crystal (Legion)
				124441, -- Leylight Shard (Legion)
				141303,141304,141305, -- Essence of Clarity (Emerald Nightmare quest item)
				139390, --Artifact Research Note
			},
		},
	} -- defaults end	

	self.version = GetAddOnMetadata("SFLoot", "Version");	
	
	self.options = {
		name = "SFLootMaster",
		type = "group",
		handler = self,
		get = "DBGet",
		set = "DBSet",
		args = {
			settings = {
				order = 1,
				type = "group",
				name = "Einstellungen",
				childGroups = "tab",
				args = {
					version = {
						order = 1,
						type = "description",
						name = "|cFF87CEFAv: "..self.version.."|r",
					},
					generalSettingsTab = {
						order = 2,
						type = "group",
						name = "Allgemein",
						childGroups = "tab",
						args = {
							generalOptions = {
								order = 1,
								name = "Loot-Fenster Öffnen",
								type = "group",
								inline = true,
								args = {
									lf_always = {
										order = 1,
										name = "Immer",
										desc = "Loot-Fenster am Anfang einer Loot-Session immer öffnen.",
										type = "toggle",
									},
									lf_main = {
										order = 2,
										name = "Main",
										desc = "Loot-Fenster bei Main Wurf öffnen",
										type = "toggle",
									},
									lf_off = {
										order = 3,
										name = "Off",
										desc = "Loot-Fenster bei Off Wurf öffnen",
										type = "toggle",
									},
									lf_transmog = {
										order = 4,
										name = "Transmog",
										desc = "Loot-Fenster bei Transmog Wurf öffnen",
										type = "toggle",
									},
									lf_pass = {
										order = 5,
										name = "Passen",
										desc = "Loot-Fenster bei Passen öffnen",
										type = "toggle",
									},									
									auto_close = {
										order = 6,
										name = "Automatisch schließen",
										desc = "Loot-Fenster automatisch am Ende der Loot-Session schließen",
										type = "toggle",
									},
								},
							},
							windowOptions = {
								order = 2,
								name = "Loot-Fenster Skalierung",
								type = "group",
								inline = true,
								args = {
									frame_scale = {
										order = 1,
										name = "Skalierung",
										desc = "Loot-Fenster skalierung",
										type = "range",
										width = "full",
										min  = 0.50,
										max  = 1.00,
										step = 0.05,
										set  = function(_, v) 
											db.frame_scale = v
											if frames.main then
												frames.main:SetScale(v)
											end
										end,
									},
								},
							},
							rollOptions = {
								order = 3,
								name = "Roll-Fenster Einstellungen",
								type = "group",
								inline = true,
								args = {
									loot_frame_direction = {
										order = 1,
										name = "Nach unten erweitern",
										desc = "Roll-Fenster nach Unten erweitern\n(Standard: Nach Oben erweitern)",
										type = "toggle",
										set = function()
											db.loot_frame_direction = not db.loot_frame_direction
											SFLoot:UpdateLootFrames()
										end,
									},
									show_loot_anchor = {
										order = 2,
										name = "Frame-Anker anzeigen",
										desc = "Den Frame-Anker für die Roll-Fenster anzeigen",
										type = "toggle",
										set = function()
											db.show_loot_anchor = not db.show_loot_anchor
											if db.show_loot_anchor then
												frames.lootFrameAnchor:Show()
											else
												frames.lootFrameAnchor:Hide()
											end
										end,
									},									
								},
							},										
							mlOptions = {
								order = 4,
								name = "Master Looter Einstellungen",
								type = "group",
								inline = true,
								args = {
									change_loot_method = {
										order = 1,
										name = "Loot-Methode",
										desc = "Loot-Methode automatisch auf Master Looter setzen wenn man der Gruppenleiter ist",
										type = "toggle",
									},
									--[[secure_roll = {
										order = 2,
										name = "Sicheres Würfeln",
										desc = "Würfelergebnisse werden vom Master Looter generiert um Manipulationen zu verhindern.",
										type = "toggle",
									},]]
								},								
							},
						},
					},
				},
			},
		},
	}
	
	self.options.args.settings.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("SFLootMaster", self.options)

	-- add it to blizz options
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SFLootMaster", "SFLootMaster", nil, "settings")
		
	menuFrame = CreateFrame("Frame", "SFLoot_Frame_RightclickMenu", UIParent, "Lib_UIDropDownMenuTemplate");
	Lib_UIDropDownMenu_Initialize(menuFrame, self.RightClickMenu, "MENU")
	
	self.db = LibStub("AceDB-3.0"):New("SFLootDB", self.defaults, true)
	db = self.db.profile;
	
	self:RegisterChatCommand("sfl", "ChatCommand")
	
	self:LocalizeSubTypes()
	
	self:CreateLootFrameAnchor();
end

function SFLoot:OnEnable()
	self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED", "OnEvent")
	self:RegisterEvent("RAID_INSTANCE_WELCOME", "OnEvent")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "OnEvent")
	self:RegisterEvent("GROUP_ROSTER_UPDATE", "OnEvent")
	
	self:RegisterComm("SFLoot");
	
	self.frames = frames
	
	self.realmName = select(2, UnitFullName("player"))
	self.playerName = self:UnitName("player")
	self.playerClass = select(2, UnitClass("player"))

	self.running = false;
	self.enabled = false;
	self.lootOpen = false;
	
	self.newVersion = false;
	
	if IsInRaid() then
		self:ScheduleTimer("GetMLDelay", 5)
	end
end

function SFLoot:GetMLDelay()
	self.isMasterLooter, self.masterLooter = self:GetML();
	self:Debug("GETML", self.isMasterLooter, self.masterLooter)

	if self.masterLooter then
		self.enabled = true
	end	

	self:CheckForActiveSession()
	
	self:RegisterMLEvents();
end

function SFLoot:CheckForActiveSession()
	local instance_type = select(2, GetInstanceInfo())
	local unit
	
	self:Debug("CheckForActiveSession")
	
	if instance_type == "raid" then
		if self.isMasterLooter then
			for i = 1, GetNumGroupMembers() do
				unit = "raid"..i
				self:Debug(unit, UnitExists(unit), UnitIsConnected(unit), self:UnitName(unit))
				if UnitExists(unit) and UnitIsConnected(unit) and not self:UnitIsUnit(unit, "player") then
					self:SendCommand(self:UnitName(unit), "ml_request_loot")
					break
				end
			end
		else
			self:SendCommand(self.masterLooter, "request_loot")
		end
	end
end

function SFLoot:RegisterMLEvents()
	if self.isMasterLooter then
		self:RegisterEvent("LOOT_OPENED", "OnEvent")
		self:RegisterEvent("LOOT_CLOSED", "OnEvent")
	else
		self:UnregisterEvent("LOOT_OPENED", "OnEvent")
		self:UnregisterEvent("LOOT_CLOSED", "OnEvent")	
	end
end

function SFLoot:OnEvent(event, ...)
	if event == "LOOT_OPENED" then 
		self.lootOpen = true
		self:LootOpened()
	elseif event == "LOOT_CLOSED" then
		self.lootOpen = false
	elseif event == "PARTY_LOOT_METHOD_CHANGED" then
		self:NewMLCheck()
	elseif event == "RAID_INSTANCE_WELCOME" then
		self:ScheduleTimer("OnRaidEnter", 2)
	elseif event == "GROUP_ROSTER_UPDATE" then
		self.isMasterLooter, self.masterLooter = self:GetML()
	elseif event == "PLAYER_ENTERING_WORLD" then
		if IsInGuild("player") then
			self:SendCommand("guild", "version", self.version)
			self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		end
	end
end

function SFLoot:OnRaidEnter()
	local instance_type = select(2, GetInstanceInfo())
	
	self.isMasterLooter, self.masterLooter = self:GetML()
	
	if db.change_loot_method and instance_type == "raid" and not self.masterLooter and UnitIsGroupLeader("player") then
		SetLootMethod("master", UnitName("player"))
		self:Print(" SFLoot aktiv. Du bist jetzt der Master Looter.")
		self:ScheduleTimer("SetLootThreshold", 1) -- if you call SetLootThreshold here LootMethod will reset to personal loot. Blizzard bug?
		self.isMasterLooter, self.masterLooter = true, self.playerName
	end
end

function SFLoot:SetLootThreshold()
	SetLootThreshold(4)
end

function SFLoot:GetML()
	local lootMethod, mlPartyID, mlRaidID = GetLootMethod()
	if lootMethod == "master" then
		local name;
		if mlRaidID then 				-- Someone in raid
			name = self:UnitName("raid"..mlRaidID)
		elseif mlPartyID == 0 then -- Player in party
			name = self.playerName
		elseif mlPartyID then		-- Someone in party
			name = self:UnitName("party"..mlPartyID)
		end
		
		if name then self.enabled = true end
		
		return IsMasterLooter(), name
	end
	
	self.enabled = false
	
	return false, nil;
end

function SFLoot:NewMLCheck()
	self.isMasterLooter, self.masterLooter = self:GetML()
	self:RegisterMLEvents()
	if self.masterLooter and self.masterLooter ~= "" and strfind(self.masterLooter, "Unknown") then
		self:Debug("Unknown ML")
		return self:ScheduleTimer("NewMLCheck", 2)
	end
end

function SFLoot:ChatCommand(msg)
	local input, arg1, arg2 = self:GetArgs(msg,3)
	input = strlower(input or "")
	self:Debug("/", input, arg1, arg2)
	if not input or input:trim() == "" or input == "help" then
		self:Print(format("Version: %s", self.version))
		self:Print("Chat commands:")
		self:Print("anchor oder a - Zeigt den Loot Frame Anchor.")
		self:Print("config oder c - Öffnet die Einstellungen.")
		self:Print("show oder s - Öffnet das Loot-Fenster bei aktiver Loot-Session.")
		self:Print("test oder t - Öffnet das Loot-Fenster mit Test-Items")
		self:Debug("- debug or d - Toggle debugging")
	elseif input == "anchor" or input == "a" then
		frames.lootFrameAnchor:Show()
		db.show_loot_anchor = true
	elseif input == "config" or input == "c" then
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)		
	elseif input == "debug" or input == "d" then
		db.debug = not db.debug
		self:Print("Debug = "..tostring(db.debug))
	elseif input == "show" or input == "s" then
		if #lootTable > 0 then
			self.MainFrameShow()
		end
	elseif input == "test" or input == "t" then
		self:Test()

	elseif input == "request" or input == "r" then
		self:SendCommand(self.masterLooter, "request_loot")
	else
		self:ChatCommand("help")
	end
end

function SFLoot:Test()	
	if self.running then return end
	testing= true
	
	lootTable = {}
	
	local test_items = {
		"\124cffa335ee\124Hitem:142519::::::::110::::2:1497:3468:\124h[Favor of the Prime Designate]\124h\124r",
		"\124cffa335ee\124Hitem:142425::::::::110::::1:3467:\124h[Goldrune Legplates]\124h\124r",
		"\124cffa335ee\124Hitem:142520::::::::110::::1:3467:\124h[Ring of Ascended Glory]\124h\124r",
		"\124cffa335ee\124Hitem:142508::::::::110::::1:3467:\124h[Chains of the Valorous]\124h\124r",
		"\124cffa335ee\124Hitem:142540::::::::110::::1:3467:\124h[Mantle of the Victorious Dead]\124h\124r",
		"\124cffa335ee\124Hitem:142512::::::::110::::1:3467:\124h[Accursed Cuspid]\124h\124r",
		--"\124cffff8000\124Hitem:19019::::::::110:::::\124h[Thunderfury, Blessed Blade of the Windseeker]\124h\124r",
	}
	
	local test_table = {}
	
	for i = 1, #test_items do
		table.insert(test_table, {
			index = i,
			link = test_items[i],
		})
	end
	
	self:BuildLootTable(test_table)
end

function SFLoot:LootOpened()
	if self.isMasterLooter and GetNumLootItems() > 0 then
		local updatedLootSlot = {}

		if not self.running then 
			mlLoot = {};
		end
		
		for i = 1, GetNumLootItems() do
			local item = GetLootSlotLink(i)
			
			if not item then
				self:ScheduleTimer("LootOpened", 1)
				return
			end
			
			-- We have reopened the loot frame, so check if we should update .lootSlot
			if self.running then
				for session = 1, #mlLoot do
					-- Just skip if we've already awarded the item or found a fitting lootSlot
					if not updatedLootSlot[session] then
						if item == mlLoot[session].link then
							if i ~= mlLoot[session].lootSlot then -- It has changed!
								mlLoot[session].lootSlot = i -- and update it
								updatedLootSlot[session] = true
								self:Debug("Changed lootSlot", session, i)
							end
							break
						end
					end
				end		
			else
				local _, _, quantity, quality = GetLootSlotInfo(i)
				local itemid = self:GetItemIDFromLink(item);	
				
				if not db.ignore[itemid] and quantity and quality and quantity > 0 and quality > 3 then 
					self:MLAddItem(item, i)
				end
			end
		end
		if #mlLoot > 0 and not self.running then
			self:Debug("session started")
			self:StartSession()
		end
	end
end

function SFLoot:MLAddItem(link, lootSlot)
	table.insert(mlLoot, {
			index = (#mlLoot + 1),
			link = link,
			lootSlot = lootSlot,
	})
	
	self:Debug("MLAddItem: "..tostring((#mlLoot + 1))..", ", link)
end

function SFLoot:AwardItem(winner, session)
	if not mlLoot[session].lootSlot then
		self:Debug("Session "..session.." didn't have lootSlot")
		return
	end

	if winner then
		local awarded = false
		if not self.lootOpen then -- we can't give out loot without the loot window open
			self:Print("Das Loot-Fenster muss offen sein während Items vergeben werden!")
			return
		end
		for i = 1, MAX_RAID_MEMBERS do
			if self:UnitIsUnit(GetMasterLootCandidate(mlLoot[session].lootSlot, i), winner) then
				GiveMasterLoot(mlLoot[session].lootSlot, i)
				awarded = true
			end
		end

		if awarded then
			self:SendCommand("group", "awarded", winner, session)
		else 
			self:Print(format("Kann %s nicht an %s vergebben - (Offline, nicht in Gruppe oder Instanz?)", lootTable[session].link, winner))
		end
	end
end

function SFLoot:BuildLootTable(item_table)
	local name, link, rarity, ilvl, iMinLevel, type, subType, iStackCount, equipLoc, texture
	
	if #item_table == 0 then return end
		
	lootTable = {}
	currentItem = 1
	
	for i = 1, #item_table do
		name, link, rarity, ilvl, iMinLevel, type, subType, iStackCount, equipLoc, texture = GetItemInfo(item_table[i].link)
		
		if not name then
			self:ScheduleTimer("BuildLootTable", 1, item_table)
			self:Debug("item not cached")
			break
		end
		
		table.insert(lootTable, {
			index      = item_table[i].index,
			name       = name,
			link       = link,
			quality    = quality,
			ilvl       = ilvl,
			subType    = subType,
			equipLoc   = equipLoc,
			texture    = texture,
			boe        = SFLoot:IsItemBoE(link),
			candidates = {},
		})		
end

	if db.lf_always or self.isMasterLooter then
		self:MainFrameShow()
	end	
	
	self:UpdateLootFrames()
	
	self:Debug("UpdateLootTable-> NewItemCount:", #lootTable)
end

function SFLoot:StartSession()
	self:SendCommand("group", "lootTable", mlLoot)
	requested_rolls = {}
end

function SFLoot:EndSession()
	self:SendCommand("group", "session_end")	
	self:CancelAllTimers()
end

function SFLoot:HaveAllItemsBeenAwarded()
	local moreItems = true
	for i = 1, #lootTable do
		if not lootTable[i].awarded then
			moreItems = false
		end
	end
	return moreItems
end

function SFLoot:LootRoll(response, s, r)
	if #lootTable == 0 then return end

	local roll = {};
	local item = lootTable[s]
	local g1, g2;
	
	if not item or not r then return end
	
	if item.equipLoc == "" and item.subType == self.db.global.localizedSubTypes["Artifact Relic"] then
		g1, g2 = self:GetArtifactRelics(item.link)
	else
	 	g1, g2 = self:GetPlayersGear(item.link, item.equipLoc)
	end
	local diff = 0
	if g2 then
		local g1diff, g2diff = select(4, GetItemInfo(g1)), select(4, GetItemInfo(g2))
		diff = g1diff >= g2diff and item.ilvl - g2diff or item.ilvl - g1diff
	elseif g1 then -- Artifact Relic might be nil
		diff = (item.ilvl - select(4, GetItemInfo(g1)))
	end	
	
	roll = {
		player = self.playerName,
		class = self.playerClass,
		index = item.index,
		lootSlot = item.lootSlot,
		response = response,
		ilvl = select(2, GetAverageItemLevel()),
		gear1 = g1,
		gear2 = g2,
		diff = diff,
		value = r,
	}
	
	local show = false
	
	if db.lf_always then
		show = true	
	elseif db.lf_main and response == 4 then
		show = true
	elseif db.lf_off and response == 3 then
		show = true
	elseif db.lf_transmog and response == 2 then
		show = true
	elseif db.lf_pass and response == 1 then
		show = true
	end
	
	if show then
		self:MainFrameShow();
		self:SwitchItem(item.index)
	end
	
	if testing then
		self:HandleLootRoll(roll)
	else
		self:SendCommand("group", "lootRoll", roll)
	end
end

function SFLoot:HandleLootRoll(roll)
	local t = lootTable[roll.index];
	
	if not t or not self.running and not testing then return end
	
	if self.isMasterLooter and not testing then
		local playerName = self.Ambiguate(roll.player)
		local token = format("%s%s", playerName, roll.index)
		
		if not requested_rolls[token] then
			self:Print("Würfelergebnis ohne vorherige Anfrage von:", playerName, "wird ignoriert!")
			return
		else
			if requested_rolls[token] ~= roll.value then
				self:Print("Würfelergebnis manipuliert von:", playerName, "wird ignoriert!")
				return	
			end
		end
	end
	
	if t.candidates[roll.player] and roll.response > 1 then 
		return
	elseif t.candidates[roll.player] and roll.response == 1 then
		t.candidates[roll.player].response = roll.response;
	elseif not t.candidates[roll.player] then
		t.candidates[roll.player] = {
			class = roll.class,
			response = roll.response,
			ilvl = roll.ilvl,
			diff = roll.diff,
			gear1 = roll.gear1,
			gear2 = roll.gear2,
			roll = roll.value,
		}
	end
	
	if roll.player == self.playerName then
		self:UpdateLootFrames()
	end
	
	if frames.main and frames.main:IsVisible() and roll.index == currentItem then
		self:SwitchItem(currentItem);
	end
end

function SFLoot:ItemAwarded(name, index)
	local item = lootTable[index]
		
	item.awarded = SFLoot.Ambiguate(name)
	
	self:UpdateItemDB(name, item)
	
	self:Print(format("%s an %s vergeben.", item.link, item.awarded))

	if self.isMasterLooter and self:HaveAllItemsBeenAwarded() then 
		self:EndSession()
		mlLoot = {}
	end
	
	self:UpdateLootFrames()
	
	if frames.main and frames.main:IsVisible() and index == currentItem then
		self:SwitchItem(currentItem)
	end
end

function SFLoot:UpdateItemDB(name, item)
	today = date("%Y/%m/%d", time());
	if not db.lootDB[today] then
		db.lootDB = {};
		db.lootDB[today] = {};
	end
	
	if not db.lootDB[today][name] then
		db.lootDB[today][name] = {};
	end
	
	table.insert(db.lootDB[today][name], item);
end

function SFLoot:GetItemDBNumItems(name)
	if not db.lootDB[today] then 
		return 0;
	elseif not db.lootDB[today][name] then
		return 0;
	else
		return #db.lootDB[today][name];
	end
end

function SFLoot:SendCommand(target, command, ...)
	-- send all data as a table, and let receiver unpack it
	local toSend = self:Serialize(command, {...})

	if target == "group" then
		if GetNumGroupMembers() > 0 then -- SendAddonMessage auto converts it to party if needed
			self:SendCommMessage("SFLoot", toSend, "RAID")
		--[[elseif num > 0 then -- Party
			self:SendCommMessage("RCLootCouncil", toSend, "PARTY")]]
		else--if self.testMode then -- Alone (testing)
			self:SendCommMessage("SFLoot", toSend, "WHISPER", self.playerName)
		end

	elseif target == "guild" then
		self:SendCommMessage("SFLoot", toSend, "GUILD")

	elseif target then
		if self:UnitIsUnit(target,"player") then -- If target == "player"
			self:SendCommMessage("SFLoot", toSend, "WHISPER", self.playerName)
		else
			-- We cannot send "WHISPER" to a crossrealm player
			if target:find("-") then
				if target:find(self.realmName) then -- Our own realm, just send it
					self:SendCommMessage("SFLoot", toSend, "WHISPER", target)
				else
					return
				end
			else -- Should also be our own realm
				self:SendCommMessage("SFLoot", toSend, "WHISPER", target)
			end
		end
	end
end

function SFLoot:OnCommReceived(prefix, serializedMsg, distri, sender)
	if prefix == "SFLoot" then
		local test, command, data = self:Deserialize(serializedMsg)

		if test then
			if command == "lootTable" then
				if self:UnitIsUnit(sender, self.masterLooter) and self.enabled then
					if testing then testing = false end
					local item_table = unpack(data)
					self:SendCommand("group", "lootAck", self.playerName) -- send ack
					self.running = true
					if frames.main and frames.main:IsVisible() then frames.main:Hide() end
					self:BuildLootTable(item_table)
				else
					self:Debug("non-ML or not enabled", sender)
				end
			elseif command == "request_loot" and self.isMasterLooter and self.running then
				self:Debug(sender, "requested mlLoot. sending...")
				self:SendCommand(sender, "requested_loot", lootTable)
			elseif command == "requested_loot" then
				lootTable = unpack(data)
				self.running = true
				self:Print("Loot Table wurde vom Master Looter empfangen.")
				self:MainFrameShow();
				self:UpdateLootFrames();
			elseif command == "ml_request_loot" then
				if self.running then
					self:SendCommand(self.masterLooter, "ml_requested_loot", lootTable)
				end
			elseif command == "ml_requested_loot" then
				if self.isMasterLooter then
					lootTable = unpack(data)
					mlLoot = {}
					self.running = true
					for i = 1, #lootTable or 0 do
						table.insert(mlLoot, {
							link = lootTable[i].link,
							lootSlot = i,
						})
					end
					self:MainFrameShow()
					self:UpdateLootFrames()
					self:Print("Loot Table wurde vom Raid angefordert.")
				end
			elseif command == "lootRoll" then
				local roll = unpack(data);
				self:HandleLootRoll(roll)
			elseif command == "request_roll" then
				self:Debug("comm: request_roll")
				if self.isMasterLooter then
					local response, s = unpack(data)
					local token = format("%s%s", sender, s)
					self:Debug(sender, "request_roll", response, s, token)
					if not requested_rolls[token] then
					    local roll = math.random(100)
						self:SendCommand(sender, "requested_roll", response, s, roll)
						requested_rolls[token] = roll
					end
				end
			elseif command == "requested_roll" then
				if self:UnitIsUnit(sender, self.masterLooter) then
					local response, s, roll = unpack(data)
					self:Debug(response, s, roll)
					self:LootRoll(response, s, roll)
				end
			elseif command == "awarded" then
				local name, index = unpack(data);
				self:ItemAwarded(name, index);
			elseif command == "session_end" then
				self:Debug("Loot Session ended.")
				if db.auto_close then frames.main:Hide() end
				self:Print("Alle Items wurden vergeben.")
				self.running = false
				secure_roll = false
				self:UpdateLootFrames()
			elseif command == "change_response" then
				local item, candidateName, response = unpack(data);
				if lootTable[item] and lootTable[item].candidates[candidateName] then
					lootTable[item].candidates[candidateName].response = response;
					if item == currentItem then
						self:SwitchItem(currentItem)
					end
				end
			elseif command == "version" and not self:UnitIsUnit(sender, "player") then
				local s_version = unpack(data);
				if (strfind(s_version, "beta")) then	
					return
				elseif self.version < s_version and not self.newVersion then
					self:Print("Eine neuere Version ist verfügbar:", s_version)
					self.newVersion = true;
				end
			end
		end
	end
end

function SFLoot:SwitchItem(s)
	-- Start with setting up some statics
	local old = currentItem
	currentItem = s
	local t = lootTable[s] -- Shortcut
	
	if not t then return end
	
	frames.main.itemIcon:SetNormalTexture(t.texture)
	frames.main.itemText:SetText(t.link)
	frames.main.iState:SetText(self:GetItemStatus(t.link))
	frames.main.itemLvl:SetText(format("ilvl: %d", t.ilvl))
	-- Set a proper item type text
	if t.subType and t.subType ~= "Miscellaneous" and t.subType ~= "Junk" and t.equipLoc ~= "" then
		frames.main.itemType:SetText(getglobal(t.equipLoc)..", "..t.subType); -- getGlobal to translate from global constant to localized name
	elseif t.subType ~= "Miscellaneous" and t.subType ~= "Junk" then
		frames.main.itemType:SetText(t.subType)
	else
		frames.main.itemType:SetText(getglobal(t.equipLoc));
	end
	
	if t.awarded then
		frames.main.awardString:SetText("Item vergeben an: "..t.awarded);
		frames.main.awardString:SetTextColor(0, 1, 0, 1) 
	else
		frames.main.awardString:SetText("Item nocht nicht vergeben");
		frames.main.awardString:SetTextColor(1, 1, 0, 1) 
	end
	
	--[[if t.candidates[self.playerName] then
		frames.main.mainBtn:Disable();
		frames.main.offBtn:Disable();
		frames.main.transmogBtn:Disable();
		if t.candidates[self.playerName].response == 1 then
			frames.main.passBtn:Disable();
		else
			frames.main.passBtn:Enable();
		end
	else
		frames.main.mainBtn:Enable();
		frames.main.offBtn:Enable();
		frames.main.transmogBtn:Enable();
		frames.main.passBtn:Enable();
	end]]
	frames.main.passBtn:Disable();
	
	if t.candidates[self.playerName] and t.candidates[self.playerName].response ~= 1 then
		frames.main.passBtn:Enable();
	end
	
	-- Update the session buttons
	
	frames.itemButtons[s] = self:CreateOrUpdateItemButton(s, t.texture, t.link, t.awarded)
	
	if old and lootTable[old] then
		frames.itemButtons[old] = self:CreateOrUpdateItemButton(old, lootTable[old].texture, lootTable[old].link, lootTable[old].awarded)
	end

	-- Since we switched sessions, we want to sort by response
	self:BuildST()
	frames.main.st.cols[3].sort = "desc"
	frames.main.st:SortData();
end

function SFLoot:MainFrameShow()
	SFLoot:CreateRollFrame();
	SFLoot.frames.main:SetScale(db.frame_scale)
	SFLoot:UpdateItemButtons();
	SFLoot:BuildST();
	SFLoot.frames.main:Show();
end

function SFLoot:UpdateItemButtons()
	if #lootTable == 0 then return end
	
	local item;
	
	for i = 1, #lootTable do
		item = lootTable[i];
		
		if item then
			frames.itemButtons[i] = self:CreateOrUpdateItemButton(i, item.texture, item.link, item.awarded)
			frames.itemButtons[i]:Show()
		end
	end
		
	for i = 1, #frames.itemButtons do
		if not lootTable[i] then
			frames.itemButtons[i]:Hide()
		end
	end
	
	self:SwitchItem(1);
end

function SFLoot:UpdateLootFrames()
	if #lootTable == 0 then return end
	
	local item

	for i = 1, #frames.lootFrames do
		frames.lootFrames[i]:Hide();
	end	
	
	if not self.running and not testing then return end
	
	local nextFrame = 1
	
	local numLootTable = #lootTable
	
	for i = 1, numLootTable do
		if db.loot_frame_direction then
			item = lootTable[i]
		else
			item = lootTable[((numLootTable + 1) - i)]
		end
		
		if item and not item.candidates[self.playerName] and not item.awarded then
			frames.lootFrames[nextFrame] = self:CreateOrUpdateLootFrames(nextFrame, item)
			frames.lootFrames[nextFrame]:Show();
			--self:Debug("Frame: "..tostring(nextFrame), "Item: "..tostring(item.index))
			nextFrame = nextFrame + 1
		end
	end
end

function SFLoot:CreateOrUpdateLootFrames(nextFrame, item)
	local lootFrames = frames.lootFrames;
	local f = self:CreateLootFrame(nextFrame);
	local p1, p2, y
	
	if db.loot_frame_direction then --true = down
		p1 = "TOP"
		p2 = "BOTTOM"
		y = -3
	else --false = up
		p1 = "BOTTOM"
		p2 = "TOP"
		y = 3
	end
	
	f:ClearAllPoints()
	
	if nextFrame == 1 then
		f:SetPoint(p1, frames.lootFrameAnchor, p2, 0 , y)
	else
		f:SetPoint(p1, lootFrames[nextFrame-1], p2, 0, y)
	end
	
	if strlen(item.name) > 30 then
		f:SetWidth(strlen(item.name)*11)
	else
		f:SetWidth(350)
	end
	
	f:SetScale(0.85)
	
	f.itemIcon:SetNormalTexture(item.texture)
	
	f.itemIcon:SetScript("OnEnter", function()
		if not lootTable then return; end
		SFLoot:CreateHypertip(lootTable[item.index].link)
	end)
	f.itemIcon:SetScript("OnLeave", SFLoot.HideTooltip)
	f.itemIcon:SetScript("OnClick", function()
		if not lootTable then return; end
		if ( IsModifiedClick() ) then
			HandleModifiedItemClick(lootTable[item.index].link);
		end
	end);
	
	f.itemText:SetText(item.link)

	f.iState:SetText(self:GetItemStatus(item.link))
	f.itemLvl:SetText(format("ilvl: %d", item.ilvl))	
	
	if item.subType and item.subType ~= "Miscellaneous" and item.subType ~= "Junk" and item.equipLoc ~= "" then
		f.itemType:SetText(getglobal(item.equipLoc)..", "..item.subType); -- getGlobal to translate from global constant to localized name
	elseif item.subType ~= "Miscellaneous" and item.subType ~= "Junk" then
		f.itemType:SetText(item.subType)
	else
		f.itemType:SetText(getglobal(item.equipLoc));
	end	
	
	f.mainBtn:SetScript("OnClick", function()
		if testing then
			SFLoot:LootRoll(4, item.index, math.random(100))
		else
			SFLoot:SendCommand(SFLoot.masterLooter, "request_roll", 4, item.index)
		end
	end)

	f.offBtn:SetScript("OnClick", function()
		if testing then
			SFLoot:LootRoll(3, item.index, math.random(100))
		else
			SFLoot:SendCommand(SFLoot.masterLooter, "request_roll", 3, item.index)
		end
	end)

	f.transmogBtn:SetScript("OnClick", function()
		if testing then
			SFLoot:LootRoll(2, item.index, math.random(100))
		else
			SFLoot:SendCommand(SFLoot.masterLooter, "request_roll", 2, item.index)
		end
	end)
	
	f.passIcon:SetScript("OnClick", function(self, button, down)
		if button == "LeftButton" and not down then
			if testing then
				SFLoot:LootRoll(1, item.index, math.random(100))
			else
				SFLoot:SendCommand(SFLoot.masterLooter, "request_roll", 1, item.index)
			end
		end
	end);
	
	return f
end

function SFLoot:CreateLootFrame(i)
	local frame = frames.lootFrames[i]
	
	if not frame then
		frame = CreateFrame("Frame", "SFLootFrame"..i, UIParent)
		frame:Hide()
		frame:SetFrameStrata("DIALOG")
		frame:SetWidth(350)
		frame:SetHeight(70)
		frame:SetBackdrop({
		bgFile = db.UI.default.background,
			edgeFile = db.UI.default.border,	   tile = true, tileSize = 64, edgeSize = 12,
			insets = { left = 2, right = 2, top = 2, bottom = 2 }
		})
		frame:SetBackdropColor(unpack(db.UI.default.bgColor))
		frame:SetBackdropBorderColor(unpack(db.UI.default.borderColor))
		frame:SetPoint("CENTER")
		frame:SetClampedToScreen(true)
	
		local item = CreateFrame("Button", nil, frame)
		item:EnableMouse()
		item:SetNormalTexture("Interface/ICONS/INV_Misc_QuestionMark")
		item:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		item:SetSize(40, 40)
		frame.itemIcon = item

		local iTxt = frame:CreateFontString(nil, "OVERLAY", "GameFontNormalLarge")
		iTxt:SetPoint("TOPLEFT", item, "TOPRIGHT", 10, 0)
		iTxt:SetText("Fehler") -- Set text for reasons
		frame.itemText = iTxt

		local iType = frame:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		iType:SetPoint("TOPLEFT", iTxt, "BOTTOMLEFT", 0, -4)
		iType:SetTextColor(1, 1, 1) 
		iType:SetText("")
		frame.itemType = iType		

		local ilvl = frame:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		ilvl:SetPoint("TOPLEFT", iType, "TOPRIGHT", 4, 0)
		ilvl:SetTextColor(1, 1, 1) -- White
		ilvl:SetText("")
		frame.itemLvl = ilvl		

		local iState = frame:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		iState:SetPoint("LEFT", ilvl, "RIGHT", 5, 0)
		iState:SetTextColor(0,1,0,1) -- Green
		iState:SetText("")
		frame.iState = iState
		
		local b = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
		b:SetText("Main")
		b:SetSize(80,25)
		b:SetPoint("TOPLEFT", item, "BOTTOMRIGHT", 0, 4)
		frame.mainBtn = b	

		local b2 = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
		b2:SetText("Off")
		b2:SetSize(80,25)
		b2:SetPoint("TOPLEFT", b, "TOPLEFT", 90, 0)
		frame.offBtn = b2

		local b3 = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
		b3:SetText("Transmog")
		b3:SetSize(80,25)
		b3:SetPoint("TOPLEFT", b2, "TOPLEFT", 90, 0)
		frame.transmogBtn = b3	
		
		local pass = CreateFrame("Button", nil, frame)
		pass:EnableMouse()
		pass:SetNormalTexture("Interface/BUTTONS/UI-GroupLoot-Pass-Up")
		pass:SetPushedTexture("Interface/BUTTONS/UI-GROUPLOOT-PASS-DOWN")
		pass:SetHighlightTexture("Interface/BUTTONS/UI-GROUPLOOT-PASS-HIGHLIGHT", "ADD")
		pass:SetScript("OnEnter", function()  end)
		pass:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -5, -5)
		pass:SetSize(40, 40)
		frame.passIcon = pass		
	end
	
	return frame
end

function SFLoot:CreateLootFrameAnchor()
	local frame = CreateFrame("Frame", "SFLootFrameAnchor", UIParent)
	lwin:Embed(frame)
	frame:RegisterConfig(db.UI["LootFrameAnchor"])
	frame:RestorePosition()
	frame:MakeDraggable()	
	frame:SetClampedToScreen(true)
	frame:SetFrameStrata("DIALOG")
	frame:SetWidth(200)
	frame:SetHeight(60)
	frame:SetScript("OnMouseDown", function(self) self:StartMoving() end)
	frame:SetScript("OnMouseUp", function(self) 
		self:StopMovingOrSizing()
		self:SavePosition()
	end)
	frame:SetBackdrop({
	bgFile = db.UI.default.background,
		edgeFile = db.UI.default.border,	   tile = true, tileSize = 64, edgeSize = 12,
		insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})
	frame:SetBackdropColor(unpack(db.UI.default.bgColor))
	frame:SetBackdropBorderColor(unpack(db.UI.default.borderColor))
	frame:SetPoint("CENTER")

	local txt = frame:CreateFontString(nil, "OVERLAY", "GameFontNormalLarge")
	txt:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -10)
	txt:SetText("SFLoot Frame Anchor")

	local test = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
	test:SetPoint("BOTTOMLEFT", 1, 0)
	test:SetSize(80, 25)
	test:SetText("Test")
	test:SetScript("OnClick", function() 
		SFLoot:Test();
	end);	

	local grow = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
	grow:SetPoint("BOTTOMRIGHT", -1, 0)
	grow:SetSize(90, 25)
	grow:SetText("Ausrichtung")
	grow:SetScript("OnClick", function() 
		db.loot_frame_direction = not db.loot_frame_direction
		SFLoot:UpdateLootFrames()
	end);
	grow:SetScript("OnEnter", function() SFLoot:CreateTooltip("Ändert die Ausrichtung der\nRoll-Frames nach oben/unten.") end)
	grow:SetScript("OnLeave", function() SFLoot:HideTooltip() end)	
	
	local close = CreateFrame("Button", nil, frame, "UIPanelCloseButton")
	close:SetPoint("TOPRIGHT", 2, 1)
	close:SetScript("OnClick", function() 
		frame:Hide()
		db.show_loot_anchor = false
	end);
	
	frames.lootFrameAnchor = frame
	
	if db.show_loot_anchor then
		frame:Show()
	else
		frame:Hide()
	end
end

function SFLoot:CreateOrUpdateItemButton(i, texture, link, awarded)
	local sessionButtons = frames.itemButtons;
	local btn = sessionButtons[i]
	if not btn then -- create the button
		btn = CreateFrame("Button", "SFButton"..i, frames.main.content)
		btn:SetSize(40,40)
		--btn:SetText(i)
		if i == 1 then
			btn:SetPoint("TOPRIGHT", frames.main, "TOPLEFT")
		else
			btn:SetPoint("TOP", sessionButtons[i-1], "BOTTOM", 0, -2)
		end
		btn:SetScript("Onclick", function() SFLoot:SwitchItem(i); end)
		btn:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square")
		btn:GetHighlightTexture():SetBlendMode("ADD")
		btn:SetNormalTexture(texture or "Interface\\InventoryItems\\WoWUnknownItem01")
		btn:GetNormalTexture():SetDrawLayer("BACKGROUND")
	end
	-- then update it
	btn:SetNormalTexture(texture or "Interface\\InventoryItems\\WoWUnknownItem01")
	-- Set the colored border and tooltips
	btn:SetBackdrop({
		bgFile = "",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		edgeSize = 18,
		--insets = { left = -4, right = -4, top = -4, bottom = -4 }
	})
	local lines = { format("Click to switch to %s", link or "") }
	if i == currentItem then
		btn:SetBackdropBorderColor(1,1,0,1) -- yellow
		--btn:SetBackdropColor(1,1,1,1)
		btn:GetNormalTexture():SetVertexColor(1,1,1)
	elseif awarded then
		btn:SetBackdropBorderColor(0,1,0,1) -- green
		--btn:SetBackdropColor(1,1,1,0.8)
		btn:GetNormalTexture():SetVertexColor(0.8,0.8,0.8)
		tinsert(lines, "Item wurde vergeben!")
	else
		btn:SetBackdropBorderColor(1,1,1,1) -- white
		--btn:SetBackdropColor(0.5,0.5,0.5,0.8)
		btn:GetNormalTexture():SetVertexColor(0.5,0.5,0.5)
	end
	btn:SetScript("OnEnter", function() SFLoot:CreateTooltip(unpack(lines)) end)
	btn:SetScript("OnLeave", function() SFLoot:HideTooltip() end)
	--btn:SetScript("OnEnter", function() SFLoot:CreateHypertip(link) end)
	--btn:SetScript("OnLeave", function() SFLoot:HideTooltip() end)
	
	return btn;
end

function SFLoot:CreateRollFrame()
	if frames.main then return end
	
	local f = self:CreateFrame("SFLootFrame", "mainframe", "SFLoot", 450, 400);
	
	local st = LibStub("ScrollingTable"):CreateST(self.scrollCols, NUM_ROWS, ROW_HEIGHT, { ["r"] = 1.0, ["g"] = 0.9, ["b"] = 0.0, ["a"] = 0.5 }, f.content);
	st.frame:SetPoint("BOTTOMLEFT", f, "BOTTOMLEFT", 10, 10);
	st:RegisterEvents({
		["OnClick"] = function(rowFrame, cellFrame, data, cols, row, realrow, column, table, button, ...)
			if button == "RightButton" and row then
				if self.isMasterLooter then
					menuFrame.name = data[realrow].name
					Lib_ToggleDropDownMenu(1, nil, menuFrame, cellFrame, 0, 0);
				end
			end
			-- Return false to have the default OnClick handler take care of left clicks
			return false
		end,
	})	
	st:EnableSelection(true)
	f.st = st
	
	local item = CreateFrame("Button", nil, f.content)
	item:EnableMouse()
    item:SetNormalTexture("Interface/ICONS/INV_Misc_QuestionMark")
    item:SetScript("OnEnter", function()
		if not lootTable then return; end
		SFLoot:CreateHypertip(lootTable[currentItem].link)
	end)
	item:SetScript("OnLeave", SFLoot.HideTooltip)
	item:SetScript("OnClick", function()
		if not lootTable then return; end
	    if ( IsModifiedClick() ) then
		    HandleModifiedItemClick(lootTable[currentItem].link);
        end
    end);
	item:SetPoint("TOPLEFT", f, "TOPLEFT", 10, -20)
	item:SetSize(50,50)
	f.itemIcon = item

	local iTxt = f.content:CreateFontString(nil, "OVERLAY", "GameFontNormalLarge")
	iTxt:SetPoint("TOPLEFT", item, "TOPRIGHT", 10, 0)
	iTxt:SetText("Fehler") -- Set text for reasons
	f.itemText = iTxt

	local ilvl = f.content:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	ilvl:SetPoint("TOPLEFT", iTxt, "BOTTOMLEFT", 0, -4)
	ilvl:SetTextColor(1, 1, 1) -- White
	ilvl:SetText("")
	f.itemLvl = ilvl

	local iState = f.content:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	iState:SetPoint("LEFT", ilvl, "RIGHT", 5, 0)
	iState:SetTextColor(0,1,0,1) -- Green
	iState:SetText("")
	f.iState = iState

	local iType = f.content:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	iType:SetPoint("TOPLEFT", ilvl, "BOTTOMLEFT", 0, -4)
	iType:SetTextColor(0.5, 1, 1) -- Turqouise
	iType:SetText("")
	f.itemType = iType

	local awdstr = f.content:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	awdstr:SetPoint("TOPLEFT", f.content, "TOPLEFT", 15, -80)
	awdstr:SetText("Item noch nicht vergeben")
	awdstr:SetTextColor(1, 1, 0, 1) -- Yellow

	f.awardString = awdstr	
	
	--[[local b = CreateFrame("Button", nil, f.content, "UIPanelButtonTemplate")
	b:SetText("Main")
	b:SetSize(80,25)
	b:SetPoint("TOPLEFT", f, "TOPLEFT", 10, -110)
	b:SetScript("OnClick", function()
		SFLoot:LootRoll(4, currentItem);
		SFLoot:SendCommand(self.masterLooter, "request_roll", 4, currentItem)
	end)
	f.mainBtn = b	

	local b2 = CreateFrame("Button", nil, f.content, "UIPanelButtonTemplate")
	b2:SetText("Off")
	b2:SetSize(80,25)
	b2:SetPoint("TOPLEFT", b, "TOPLEFT", 90, 0)
	b2:SetScript("OnClick", function()
		--SFLoot:LootRoll(3, currentItem);
		SFLoot:SendCommand(self.masterLooter, "request_roll", 3, currentItem)
	end)
	f.offBtn = b2

	local b3 = CreateFrame("Button", nil, f.content, "UIPanelButtonTemplate")
	b3:SetText("Transmog")
	b3:SetSize(80,25)
	b3:SetPoint("TOPLEFT", b2, "TOPLEFT", 90, 0)
	b3:SetScript("OnClick", function()
		--SFLoot:LootRoll(2, currentItem);
		SFLoot:SendCommand(self.masterLooter, "request_roll", 2, currentItem)
	end)
	f.transmogBtn = b3]]

	local b4 = CreateFrame("Button", nil, f.content, "UIPanelButtonTemplate")
	b4:SetText("Passen")
	b4:SetSize(80,25)
	--b4:SetPoint("TOPLEFT", b3, "TOPLEFT", 90, 0)
	b4:SetPoint("TOPLEFT", f, "TOPLEFT", 10, -110)
	b4:SetScript("OnClick", function()
		if testing then
			SFLoot:LootRoll(1, currentItem, math.random(100));
		else
			SFLoot:SendCommand("group", "change_response", currentItem, self.playerName, 1)
		end
	end)
	f.passBtn = b4	
	
	local close = CreateFrame("Button", nil, f.content, "UIPanelCloseButton")
	close:SetPoint("TOPRIGHT", 2, 1)
	close:SetScript("OnClick", function() 
		if self.isMasterLooter and self.running then
			LibDialog:Spawn("SFLOOT_CONFIRM_ABORT")
		else
			f:Hide()
		end
		if testing then
			testing = false
			for i = 1, #frames.lootFrames do
				frames.lootFrames[i]:Hide();
			end	
		end
	end);
	
	f.closeBtn = close;
	
	f:SetWidth(st.frame:GetWidth() + 20)
	
	frames.main = f;
end

function SFLoot:BuildST()
	
	if #lootTable == 0 then return end
	
	local rows = {}
	local i = 1
	for name in pairs(lootTable[currentItem].candidates) do
		rows[i] = {
			name = name,
			cols = {
				{ value = "",	DoCellUpdate = self.SetCellClass,		name = "class",},
				{ value = "",	DoCellUpdate = self.SetCellName,			name = "name",},
				{ value = "",	DoCellUpdate = self.SetCellResponse,	name = "response",},
				{ value = "",	DoCellUpdate = self.SetCellIlvl,			name = "ilvl",},
				{ value = "",	DoCellUpdate = self.SetCellDiff,			name = "items",},
				{ value = "",	DoCellUpdate = self.SetCellGear, 		name = "gear1",},
				{ value = "",	DoCellUpdate = self.SetCellGear, 		name = "gear2",},
				{ value = "",	DoCellUpdate = self.SetCellItems,			name = "items",},
				{ value = "",	DoCellUpdate = self.SetCellRoll,			name = "roll"},
			},
		}
		i = i + 1
	end
	frames.main.st:SetData(rows)
end

function SFLoot:CreateFrame(name, cName, title, width, height)
	local f = CreateFrame("Frame", name, nil) -- LibWindow seems to work better with nil parent
	f:Hide()
	f:SetFrameStrata("DIALOG")
	f:SetWidth(width or 450)
	f:SetHeight(height or 325)
	lwin:Embed(f)
	f:RegisterConfig(db.UI[cName])
	f:RestorePosition() -- might need to move this to after whereever GetFrame() is called
	f:MakeDraggable()
	f:SetClampedToScreen(true)
	f:SetScript("OnMouseWheel", function(f,delta) if IsControlKeyDown() then lwin.OnMouseWheel(f,delta) end end)

	local tf = CreateFrame("Frame", nil, f)
	--tf:SetFrameStrata("DIALOG")
	tf:SetToplevel(true)
	tf:SetBackdrop({
	     bgFile = db.UI.default.background,
	     edgeFile = db.UI.default.border,
	     tile = true, tileSize = 64, edgeSize = 12,
	     insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})
	tf:SetBackdropColor(unpack(db.UI.default.bgColor))
	tf:SetBackdropBorderColor(unpack(db.UI.default.borderColor))
	tf:SetHeight(22)
	tf:EnableMouse()
	tf:SetMovable(true)
	tf:SetWidth((width/2) or 250)
	tf:SetPoint("CENTER",f,"TOP",0,-1)
	tf:SetScript("OnMouseDown", function(self) self:GetParent():StartMoving() end)
	tf:SetScript("OnMouseUp", function(self) -- Get double click by trapping time betweem mouse up
		local frame = self:GetParent()
		frame:StopMovingOrSizing()
		frame:SavePosition()
		if self.lastClick and GetTime() - self.lastClick <= 0.5 then
			self.lastClick = nil
			if frame.minimized then frame:Maximize() else frame:Minimize() end
		else
			self.lastClick = GetTime()
		end
	end)

	local text = tf:CreateFontString(nil,"OVERLAY","GameFontNormal")
	text:SetPoint("CENTER",tf,"CENTER")
	text:SetTextColor(1,1,1,1)
	text:SetText(title)
	tf.text = text
	f.title = tf

	local c = CreateFrame("Frame", nil, f) -- frame that contains the actual content
	c:SetBackdrop({
	bgFile = db.UI.default.background,
		edgeFile = db.UI.default.border,	   tile = true, tileSize = 64, edgeSize = 12,
		insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})
	c:EnableMouse(true)
	c:SetWidth(450)
	c:SetHeight(height or 325)
	c:SetBackdropColor(unpack(db.UI.default.bgColor))
	c:SetBackdropBorderColor(unpack(db.UI.default.borderColor))
	c:SetPoint("TOPLEFT")
	c:SetScript("OnMouseDown", function(self) self:GetParent():StartMoving() end)
	c:SetScript("OnMouseUp", function(self) self:GetParent():StopMovingOrSizing(); self:GetParent():SavePosition() end)
	f.content = c
	f.minimized = false
	f.IsMinimized = function(frame) return frame.minimized end
	f.Minimize = function(frame)
		--self:Debug("Minimize()")
		if not frame.minimized then
		  	frame.content:Hide()
			frame.minimized = true
		end
	end
	f.Maximize = function(frame)
		--self:Debug("Maximize()")
		if frame.minimized then
		  	frame.content:Show()
			frame.minimized = false
		end
	end
	-- Support for auto hide in combat:
	tinsert(frames, f)
	local old_setwidth = f.SetWidth
	f.SetWidth = function(self, width) -- Hack so we only have to set width once
		old_setwidth(self, width)
		self.content:SetWidth(width)
	end
	local old_setheight = f.SetHeight
	f.SetHeight = function(self, height)
		old_setheight(self, width)
		self.content:SetHeight(height)
	end

	return f
end

function SFLoot:GetDiffColor(num)
	if num == "" then num = 0 end -- Can't compare empty string
	local green, red, grey = {0,1,0,1},{1,0,0,1},{0.75,0.75,0.75,1}
	if num > 0 then return green end
	if num < 0 then return red end
	return grey
end

function SFLoot.SetCellClass(rowFrame, frame, data, cols, row, realrow, column, fShow, table, ...)
	local name = data[realrow].name
	SFLoot.SetCellClassIcon(rowFrame, frame, data, cols, row, realrow, column, fShow, table, lootTable[currentItem].candidates[name].class)
end

function SFLoot.SetCellName(rowFrame, frame, data, cols, row, realrow, column, fShow, table, ...)
	local name = data[realrow].name
	frame.text:SetText(SFLoot.Ambiguate(name))
	local c = SFLoot:GetClassColor(lootTable[currentItem].candidates[name].class)
	frame.text:SetTextColor(c.r, c.g, c.b, c.a)
	data[realrow].cols[column].value = name
end

function SFLoot.SetCellResponse(rowFrame, frame, data, cols, row, realrow, column, fShow, table, ...)
	local name = data[realrow].name
	frame.text:SetText(responses[lootTable[currentItem].candidates[name].response])
	frame.text:SetTextColor(unpack(response_colors[lootTable[currentItem].candidates[name].response]))
end

function SFLoot.SetCellIlvl(rowFrame, frame, data, cols, row, realrow, column, fShow, table, ...)
	local name = data[realrow].name
	frame.text:SetText(math.floor(lootTable[currentItem].candidates[name].ilvl))
	data[realrow].cols[column].value = lootTable[currentItem].candidates[name].ilvl
end

function SFLoot.SetCellDiff(rowFrame, frame, data, cols, row, realrow, column, fShow, table, ...)
	local name = data[realrow].name
	frame.text:SetText(lootTable[currentItem].candidates[name].diff);
	frame.text:SetTextColor(unpack(SFLoot:GetDiffColor(lootTable[currentItem].candidates[name].diff)))
	data[realrow].cols[column].value = lootTable[currentItem].candidates[name].diff
end

function SFLoot.SetCellGear(rowFrame, frame, data, cols, row, realrow, column, fShow, table, ...)
	local gear = data[realrow].cols[column].name -- gear1 or gear2
	local name = data[realrow].name
	gear = lootTable[currentItem].candidates[name][gear] -- Get the actual gear
	if gear then
		local texture = select(10, GetItemInfo(gear))
		frame:SetNormalTexture(texture)
		frame:SetScript("OnEnter", function() SFLoot:CreateHypertip(gear) end)
		frame:SetScript("OnLeave", function() SFLoot:HideTooltip() end)
		frame:Show()
	else
		frame:Hide()
	end
end

function SFLoot.SetCellItems(rowFrame, frame, data, cols, row, realrow, column, fShow, table, ...)
	local name = data[realrow].name
	frame.text:SetText(SFLoot:GetItemDBNumItems(name));
	data[realrow].cols[column].value = SFLoot:GetItemDBNumItems(name)
end

function SFLoot.SetCellRoll(rowFrame, frame, data, cols, row, realrow, column, fShow, table, ...)
	local name = data[realrow].name
	if lootTable[currentItem].candidates[name].response > 1 then
		frame.text:SetText(lootTable[currentItem].candidates[name].roll)
	else
		frame.text:SetText("")
	end
	data[realrow].cols[column].value = lootTable[currentItem].candidates[name].roll
end

function SFLoot.SetCellClassIcon(rowFrame, frame, data, cols, row, realrow, column, fShow, table, class)
	local celldata = data[realrow].cols and data[realrow].cols[column] or data[realrow][column]
	local class = celldata.args and celldata.args[1] or class
	if class then
		frame:SetNormalTexture("Interface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES"); -- this is the image containing all class icons
		local coords = CLASS_ICON_TCOORDS[class]; -- get the coordinates of the class icon we want
		frame:GetNormalTexture():SetTexCoord(unpack(coords)); -- cut out the region with our class icon according to coords
	else -- if there's no class
		frame:SetNormalTexture("Interface/ICONS/INV_Sigil_Thorim.png")
	end
end

--- Returns a color table for use with lib-st
function SFLoot:GetClassColor(class)
	local color = RAID_CLASS_COLORS[class]
	if not color then
		-- if class not found, return epic color.
		return {r=1,g=1,b=1,a=1}--{["r"] = 0.63921568627451, ["g"] = 0.2078431372549, ["b"] = 0.93333333333333, ["a"] = 1.0 };
	else
		color.a = 1.0
		return color
	end
end

function ResponseSort(table, rowa, rowb, sortbycol)
	local column = table.cols[sortbycol]
	local a, b = table:GetRow(rowa), table:GetRow(rowb);
	a, b = lootTable[currentItem].candidates[a.name].response, lootTable[currentItem].candidates[b.name].response
	if a == b then
		if column.sortnext then
			local nextcol = table.cols[column.sortnext];
			if not(nextcol.sort) then
				if nextcol.comparesort then
					return nextcol.comparesort(table, rowa, rowb, column.sortnext);
				else
					return table:CompareSort(rowa, rowb, column.sortnext);
				end
			end
		end
		return false
	else
		return a > b;
	end
end

function RollSort(table, rowa, rowb, sortbycol)
	local column = table.cols[sortbycol]
	local a, b = table:GetRow(rowa), table:GetRow(rowb);
	a, b = lootTable[currentItem].candidates[a.name].roll, lootTable[currentItem].candidates[b.name].roll
	if a == b then
		return false
	else
		return a > b;
	end
end

function SFLoot:UnitIsUnit(unit1, unit2)
	if not unit1 or not unit2 then return false end
	-- Remove realm names, if any
	if strfind(unit1, "-", nil, true) ~= nil then
		unit1 = Ambiguate(unit1, "short")
	end
	if strfind(unit2, "-", nil, true) ~= nil then
		unit2 = Ambiguate(unit2, "short")
	end
	return UnitIsUnit(unit1, unit2)
end

-- We always want realm name when we call UnitName
-- Note: If 'unit' is a playername, that player must be in our raid or party!
--[[ NOTE I'm concerned about the UnitIsVisible() range thing with UnitName(),
 	although testing in party doesn't seem to affect it. To counter this, it'll
	return any "unit" with something after a "-" (i.e a name from GetRaidRosterInfo())
	which means it isn't useable with all the "name-target" unitIDs]]
function SFLoot:UnitName(unit)
	-- First strip any spaces
	unit = gsub(unit, " ", "")
	-- Then see if we already have a realm name appended
	local find = strfind(unit, "-", nil, true)
	if find and find < #unit then -- "-" isn't the last character
		-- Let's give it same treatment as below so we're sure it's the same
		local name, realm = strsplit("-", unit, 2)
		name = name:lower():gsub("^%l", string.upper)
		return name.."-"..realm
	end
	-- Apparently functions like GetRaidRosterInfo() will return "real" name, while UnitName() won't
	-- always work with that (see ticket #145). We need this to be consistant, so just lowercase the unit:
	unit = unit:lower()
	-- Proceed with UnitName()
	local name, realm = UnitName(unit)
	if not realm or realm == "" then realm = self.realmName end -- Extract our own realm
	-- We also want to make sure the returned name is always title cased (it might not always be! ty Blizzard)
	name = name:lower():gsub("^%l", string.upper)
	return name and name.."-"..realm or nil
end

--- Displays a tooltip anchored to the mouse
-- @paramsig ...
-- @param ... Lines to be added.
function SFLoot:CreateTooltip(...)
	GameTooltip:SetOwner(UIParent, "ANCHOR_CURSOR")
	for i = 1, select("#", ...) do
		GameTooltip:AddLine(select(i, ...),1,1,1)
	end
	GameTooltip:Show()
end

--- Displays a hyperlink tooltip
-- @paramsig link
-- @param link The link to display
function SFLoot:CreateHypertip(link)
	if not link or link == "" then return end
	GameTooltip:SetOwner(UIParent, "ANCHOR_CURSOR")
	GameTooltip:SetHyperlink(link)
end

--- Hide the tooltip created with :CreateTooltip()
function SFLoot:HideTooltip()
	GameTooltip:Hide()
end

function SFLoot:GetItemStatus(item)
	-- addon:Debug("GetitemStatus", item)
	if not item then return "" end
	GameTooltip:SetOwner(UIParent, "ANCHOR_NONE")
	GameTooltip:SetHyperlink(item)
	local text = ""
	if GameTooltip:NumLines() > 1 then -- check that there is something here
		local line = getglobal('GameTooltipTextLeft2') -- Should always be line 2
		local t =  line:GetText()
		-- The following color string should be there if we have a green status text
		if t then
			if strfind(t, "cFF 0FF 0") then
				text = t
			end
		end
	end
	GameTooltip:Hide()
	return text
end

local INVTYPE_Slots = {
		INVTYPE_HEAD		    = "HeadSlot",
		INVTYPE_NECK		    = "NeckSlot",
		INVTYPE_SHOULDER	    = "ShoulderSlot",
		INVTYPE_CLOAK		    = "BackSlot",
		INVTYPE_CHEST		    = "ChestSlot",
		INVTYPE_WRIST		    = "WristSlot",
		INVTYPE_HAND		    = "HandsSlot",
		INVTYPE_WAIST		    = "WaistSlot",
		INVTYPE_LEGS		    = "LegsSlot",
		INVTYPE_FEET		    = "FeetSlot",
		INVTYPE_SHIELD		    = "SecondaryHandSlot",
		INVTYPE_ROBE		    = "ChestSlot",
		INVTYPE_2HWEAPON	    = {"MainHandSlot","SecondaryHandSlot"},
		INVTYPE_WEAPONMAINHAND	= "MainHandSlot",
		INVTYPE_WEAPONOFFHAND	= {"SecondaryHandSlot",["or"] = "MainHandSlot"},
		INVTYPE_WEAPON		    = {"MainHandSlot","SecondaryHandSlot"},
		INVTYPE_THROWN		    = {"SecondaryHandSlot", ["or"] = "MainHandSlot"},
		INVTYPE_RANGED		    = {"SecondaryHandSlot", ["or"] = "MainHandSlot"},
		INVTYPE_RANGEDRIGHT 	= {"SecondaryHandSlot", ["or"] = "MainHandSlot"},
		INVTYPE_FINGER		    = {"Finger0Slot","Finger1Slot"},
		INVTYPE_HOLDABLE	    = {"SecondaryHandSlot", ["or"] = "MainHandSlot"},
		INVTYPE_TRINKET		    = {"TRINKET0SLOT", "TRINKET1SLOT"}
}

local SFTokenTable = {
	
}

function SFLoot:GetPlayersGear(link, equipLoc)
	local itemID = self:GetItemIDFromLink(link) -- Convert to itemID
	--self:DebugLog("GetPlayersGear", itemID, equipLoc)
	if not itemID then return nil, nil; end
	local item1, item2;
	-- check if the item is a token, and if it is, return the matching current gear
	if SFTokenTable[itemID] then
		if SDTokenTable[itemID] == "Trinket" then -- We need to return both trinkets
			item1 = GetInventoryItemLink("player", GetInventorySlotInfo("TRINKET0SLOT"))
			item2 = GetInventoryItemLink("player", GetInventorySlotInfo("TRINKET1SLOT"))
		else	-- Just return the slot from the tokentable
			item1 = GetInventoryItemLink("player", GetInventorySlotInfo(SFTokenTable[itemID]))
		end
		return item1, item2
	end
	local slot = INVTYPE_Slots[equipLoc]
	if not slot then return nil, nil; end;
	item1 = GetInventoryItemLink("player", GetInventorySlotInfo(slot[1] or slot))
	if not item1 and slot['or'] then
		item1 = GetInventoryItemLink("player", GetInventorySlotInfo(slot['or']))
	end;
	if slot[2] then
		item2 = GetInventoryItemLink("player", GetInventorySlotInfo(slot[2]))
	end
	return item1, item2;
end

function SFLoot:GetArtifactRelics(link)
	local id = self:GetItemIDFromLink(link)
	local g1,g2;
	for i = 1, C_ArtifactUI.GetEquippedArtifactNumRelicSlots() or 0 do
		if C_ArtifactUI.CanApplyRelicItemIDToEquippedArtifactSlot(id,i) then -- We can equip it
			if g1 then
				g2 = select(4,C_ArtifactUI.GetEquippedArtifactRelicInfo(i))
			else
				g1 = select(4,C_ArtifactUI.GetEquippedArtifactRelicInfo(i))
			end
		end
	end
	return g1, g2
end

-- Used to find localized subType names
local subTypeLookup = {
	["Cloth"]					= 124168, -- Felgrease-Smudged Robes
	["Leather"] 				= 124265, -- Leggings of Eternal Terror
	["Mail"] 					= 124291, -- Eredar Fel-Chain Gloves
	["Plate"]					= 124322, -- Treads of the Defiler
	["Shields"] 				= 124354, -- Felforged Aegis
	["Bows"] 					= 128194, -- Snarlwood Recurve Bow
	["Crossbows"] 				= 124362, -- Felcrystal Impaler
	["Daggers"]					= 124367, -- Fang of the Pit
	["Guns"]						= 124370, -- Felfire Munitions Launcher
	["Fist Weapons"] 			= 124368, -- Demonblade Eviscerator
	["One-Handed Axes"]		= 128196, -- Limbcarver Hatchet
	["One-Handed Maces"]		= 124372, -- Gavel of the Eredar
	["One-Handed Swords"] 	= 124387, -- Shadowrend Talonblade
	["Polearms"] 				= 124377, -- Rune Infused Spear
	["Staves"]					= 124382, -- Edict of Argus
	["Two-Handed Axes"]		= 124360, -- Hellrender
	["Two-Handed Maces"]		= 124375, -- Maul of Tyranny
	["Two-Handed Swords"]	= 124389, -- Calamity's Edge
	["Wands"]					= 128096, -- Demonspine Wand
	["Warglaives"]				= 141604, -- Glaive of the Fallen
	["Artifact Relic"]		= 141271, -- Hope of the Forest
}

function SFLoot:LocalizeSubTypes()
	if self.db.global.localizedSubTypes.created then return end -- We only need to create it once
	-- Get the item info
	for _, item in pairs(subTypeLookup) do
		GetItemInfo(item)
	end
	self.db.global.localizedSubTypes = {} -- reset
	self.db.global.localizedSubTypes.created = true
	for name, item in pairs(subTypeLookup) do
		local sType = select(7, GetItemInfo(item))
		if sType then
			self.db.global.localizedSubTypes[sType] = name
			self:Debug("Found "..name.." localized as: "..sType)
		else -- Probably not cached, set a timer
			self:Debug("We didn't find:", name, item)
			self:ScheduleTimer("LocalizeSubTypes", 2)
			self.db.global.localizedSubTypes.created = false
			return
		end
	end
end

function SFLoot:IsItemBoE(item)
	if not item then return false end
	GameTooltip:SetOwner(UIParent, "ANCHOR_NONE")
	GameTooltip:SetHyperlink(item)
	if GameTooltip:NumLines() > 1 then -- check that there is something here
		for i = 1, 5 do -- BoE status won't be further away than line 5
			local line = getglobal('GameTooltipTextLeft' .. i)
			if line and line.GetText then
				if line:GetText() == ITEM_BIND_ON_EQUIP then
					GameTooltip:Hide()
					return true
				end
			end
		end
	end
	GameTooltip:Hide()
	return false
end

function SFLoot:GetItemIDFromLink(link)
	return tonumber(strmatch(link, "item:(%d+):"))
end

function SFLoot:GetItemStringFromLink(link)
	return strmatch(link, "item:([%d:]+)")
end

function SFLoot:GetItemNameFromLink(link)
	return strmatch(link, "%[(.+)%]")
end

function SFLoot.round(num, decimals)
	if type(num) ~= "number" then return "" end
	return tonumber(string.format("%." .. (decimals or 0) .. "f", num))
end

function SFLoot.Ambiguate(name)
	return db.ambiguate and Ambiguate(name, "none") or Ambiguate(name, "short")
end

function SFLoot:DBGet(info)
	return self.db.profile[info[#info]]
end

function SFLoot:DBSet(info, val)	
	self.db.profile[info[#info]] = val
end

function SFLoot:Debug(msg, ...)
	if db.debug then
		if select("#", ...) > 0 then
			self:Print("|cffcb6700debug:|r "..tostring(msg).."|cffff6767", ...)
		else
			self:Print("|cffcb6700debug:|r "..tostring(msg).."|r")
		end
	end
end

do
	local info -- Efficiency :)
	-- NOTE Take care of info[] values when inserting new buttons
	function SFLoot.RightClickMenu(menu, level)
		if not SFLoot.isMasterLooter then return end

		local candidateName = menu.name
		--local data = lootTable[currentItem].candidates[candidateName] -- Shorthand

		if level == 1 then
			info = Lib_UIDropDownMenu_CreateInfo()
			info.text = SFLoot.Ambiguate(candidateName)
			info.isTitle = true
			info.notCheckable = true
			info.disabled = true
			Lib_UIDropDownMenu_AddButton(info, level)

			info.text = ""
			info.isTitle = false
			Lib_UIDropDownMenu_AddButton(info, level)

			info.text = "Item zuweisen"
			info.func = function()
				SFLoot:AwardItem(candidateName, currentItem)
			end
			if lootTable[currentItem].awarded then
				info.disabled = true
			else
				info.disabled = false
			end
			Lib_UIDropDownMenu_AddButton(info, level)

			info.text = ""
			info.isTitle = false
			Lib_UIDropDownMenu_AddButton(info, level)

			info.text = "Verwendung ändern"
			info.value = "CHANGE_RESPONSE"
			if lootTable[currentItem].awarded then
				info.disabled = true
				info.hasArrow = false
			else
				info.disabled = false
				info.hasArrow = true
			end
			Lib_UIDropDownMenu_AddButton(info, level)			
		elseif level == 2 then
			local value = LIB_UIDROPDOWNMENU_MENU_VALUE
			info = Lib_UIDropDownMenu_CreateInfo()

			if value == "CHANGE_RESPONSE" then
				for i = 1, #responses do
					local v = responses[i]
					info.text = v
					info.notCheckable = true
					info.func = function()
							SFLoot:SendCommand("group", "change_response", currentItem, candidateName, i)
					end
					Lib_UIDropDownMenu_AddButton(info, level)
				end
			end
		end
	end
end

LibDialog:Register("SFLOOT_CONFIRM_ABORT", {
	text = "Es wurden noch nicht alle Items vergeben! Loot Session trotzdem beenden?",
	on_show = function(self)
		self:SetFrameStrata("FULLSCREEN")
	end,
	buttons = {
		{	text = "Ja",
			on_click = function(self)
				SFLoot:EndSession()
				frames.main:Hide();
			end,
		},
		{	text = "Nein",
		},
	},
	hide_on_escape = true,
	show_while_dead = true,
})
